var _a;
import { MAX_RECOMMENDED_TX_SIZE, cojsonInternals } from "cojson";
import { Account, CoValueBase, ItemsSym, Ref, SchemaInit, co, ensureCoValueLoaded, inspect, isRefEncoded, loadCoValue, subscribeToCoValue, subscribeToExistingCoValue, } from "../internal.js";
/** @deprecated Use CoFeed instead */
export { CoFeed as CoStream };
/**
 * CoFeeds are collaborative logs of data.
 *
 * @categoryDescription Content
 * They are similar to `CoList`s, but with a few key differences:
 * - They are append-only
 * - They consist of several internal append-only logs, one per account session (tab, device, app instance, etc.)
 * - They expose those as a per-account aggregated view (default) or a precise per-session view
 *
 * ```ts
 * favDog.push("Poodle");
 * favDog.push("Schnowzer");
 * ```
 *
 * @category CoValues
 */
export class CoFeed extends CoValueBase {
    /**
     * Declare a `CoFeed` by subclassing `CoFeed.Of(...)` and passing the item schema using a `co` primitive or a `co.ref`.
     *
     * @example
     * ```ts
     * class ColorFeed extends CoFeed.Of(co.string) {}
     * class AnimalFeed extends CoFeed.Of(co.ref(Animal)) {}
     * ```
     *
     * @category Declaration
     */
    static Of(item) {
        var _b, _c;
        return _c = class CoFeedOf extends _a {
                constructor() {
                    super(...arguments);
                    this[_b] = item;
                }
            },
            _b = co.items,
            _c;
    }
    /** @internal */
    get _schema() {
        return this.constructor._schema;
    }
    /**
     * The current account's view of this `CoFeed`
     * @category Content
     */
    get byMe() {
        if (this._loadedAs._type === "Account") {
            return this[this._loadedAs.id];
        }
        else {
            return undefined;
        }
    }
    /**
     * The current session's view of this `CoFeed`
     *
     * This is a shortcut for `this.perSession` where the session ID is the current session ID.
     *
     * @category Content
     */
    get inCurrentSession() {
        if (this._loadedAs._type === "Account") {
            return this.perSession[this._loadedAs.sessionID];
        }
        else {
            return undefined;
        }
    }
    constructor(options) {
        super();
        if (options && "fromRaw" in options) {
            Object.defineProperties(this, {
                id: {
                    value: options.fromRaw.id,
                    enumerable: false,
                },
                _raw: { value: options.fromRaw, enumerable: false },
            });
        }
        return new Proxy(this, CoStreamProxyHandler);
    }
    /**
     * Create a new `CoFeed`
     * @category Creation
     */
    static create(init, options) {
        const instance = new this({ init, owner: options.owner });
        const raw = options.owner._raw.createStream();
        Object.defineProperties(instance, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        if (init) {
            instance.push(...init);
        }
        return instance;
    }
    /**
     * Push items to this `CoFeed`
     *
     * Items are appended to the current session's log. Each session (tab, device, app instance)
     * maintains its own append-only log, which is then aggregated into the per-account view.
     *
     * @example
     * ```ts
     * // Adds items to current session's log
     * feed.push("item1", "item2");
     *
     * // View items from current session
     * console.log(feed.inCurrentSession);
     *
     * // View aggregated items from all sessions for current account
     * console.log(feed.byMe);
     * ```
     *
     * @category Content
     */
    push(...items) {
        for (const item of items) {
            this.pushItem(item);
        }
    }
    pushItem(item) {
        const itemDescriptor = this._schema[ItemsSym];
        if (itemDescriptor === "json") {
            this._raw.push(item);
        }
        else if ("encoded" in itemDescriptor) {
            this._raw.push(itemDescriptor.encoded.encode(item));
        }
        else if (isRefEncoded(itemDescriptor)) {
            this._raw.push(item.id);
        }
    }
    /**
     * Get a JSON representation of the `CoFeed`
     * @category
     */
    toJSON() {
        const itemDescriptor = this._schema[ItemsSym];
        const mapper = itemDescriptor === "json"
            ? (v) => v
            : "encoded" in itemDescriptor
                ? itemDescriptor.encoded.encode
                : (v) => v && v.id;
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(Object.entries(this).map(([account, entry]) => [
                account,
                mapper(entry.value),
            ])),
            in: Object.fromEntries(Object.entries(this.perSession).map(([session, entry]) => [
                session,
                mapper(entry.value),
            ])),
        };
    }
    /** @internal */
    [inspect]() {
        return this.toJSON();
    }
    /** @internal */
    static schema(def) {
        this._schema ||= {};
        Object.assign(this._schema, def);
    }
    /**
     * Load a `CoFeed`
     * @category Subscription & Loading
     */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /**
     * Subscribe to a `CoFeed`, when you have an ID but don't have a `CoFeed` instance yet
     * @category Subscription & Loading
     */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /**
     * Ensure a `CoFeed` is loaded to the specified depth
     *
     * @returns A new instance of the same CoFeed that's loaded to the specified depth,
     * or undefined if it cannot be loaded that deeply
     * @category Subscription & Loading
     */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /**
     * An instance method to subscribe to an existing `CoFeed`
     *
     * No need to provide an ID or Account since they're already part of the instance.
     * @category Subscription & Loading
     */
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
}
_a = CoFeed;
(() => {
    _a.prototype._type = "CoStream";
})();
/**
 * Converts a raw stream entry into a formatted CoFeed entry with proper typing and accessors.
 * @internal
 */
function entryFromRawEntry(accessFrom, rawEntry, loadedAs, accountID, itemField) {
    return {
        get value() {
            if (itemField === "json") {
                return rawEntry.value;
            }
            else if ("encoded" in itemField) {
                return itemField.encoded.decode(rawEntry.value);
            }
            else if (isRefEncoded(itemField)) {
                return this.ref?.accessFrom(accessFrom, rawEntry.by + rawEntry.tx.sessionID + rawEntry.tx.txIndex + ".value");
            }
            else {
                throw new Error("Invalid item field schema");
            }
        },
        get ref() {
            if (itemField !== "json" && isRefEncoded(itemField)) {
                const rawId = rawEntry.value;
                return new Ref(rawId, loadedAs, itemField);
            }
            else {
                return undefined;
            }
        },
        get by() {
            return (accountID &&
                new Ref(accountID, loadedAs, {
                    ref: Account,
                    optional: false,
                })?.accessFrom(accessFrom, rawEntry.by + rawEntry.tx.sessionID + rawEntry.tx.txIndex + ".by"));
        },
        madeAt: rawEntry.at,
        tx: rawEntry.tx,
    };
}
/**
 * The proxy handler for `CoFeed` instances
 * @internal
 */
export const CoStreamProxyHandler = {
    get(target, key, receiver) {
        if (typeof key === "string" && key.startsWith("co_")) {
            const rawEntry = target._raw.lastItemBy(key);
            if (!rawEntry)
                return;
            const entry = entryFromRawEntry(receiver, rawEntry, target._loadedAs, key, target._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = target._raw.itemsBy(key);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(receiver, rawEntry.value, target._loadedAs, key, target._schema[ItemsSym]);
                        }
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    })();
                },
            });
            return entry;
        }
        else if (key === "perSession") {
            return new Proxy({}, CoStreamPerSessionProxyHandler(target, receiver));
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        if (key === ItemsSym && typeof value === "object" && SchemaInit in value) {
            target.constructor._schema ||= {};
            target.constructor._schema[ItemsSym] =
                value[SchemaInit];
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        if (descriptor.value &&
            key === ItemsSym &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            target.constructor._schema ||= {};
            target.constructor._schema[ItemsSym] =
                descriptor.value[SchemaInit];
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
    ownKeys(target) {
        const keys = Reflect.ownKeys(target);
        for (const accountID of target._raw.accounts()) {
            keys.push(accountID);
        }
        return keys;
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
};
/**
 * The proxy handler for the per-session view of a `CoFeed`
 * @internal
 */
const CoStreamPerSessionProxyHandler = (innerTarget, accessFrom) => ({
    get(_target, key, receiver) {
        if (typeof key === "string" && key.includes("session")) {
            const sessionID = key;
            const rawEntry = innerTarget._raw.lastItemIn(sessionID);
            if (!rawEntry)
                return;
            const by = cojsonInternals.accountOrAgentIDfromSessionID(sessionID);
            const entry = entryFromRawEntry(accessFrom, rawEntry, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                ? by
                : undefined, innerTarget._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = innerTarget._raw.itemsIn(sessionID);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(accessFrom, rawEntry.value, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                                ? by
                                : undefined, innerTarget._schema[ItemsSym]);
                        }
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    })();
                },
            });
            return entry;
        }
        else {
            return Reflect.get(innerTarget, key, receiver);
        }
    },
    ownKeys() {
        return innerTarget._raw.sessions();
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
});
/** @deprecated Use FileStream instead */
export { FileStream as BinaryCoStream };
/**
 * FileStreams are `CoFeed`s that contain binary data, collaborative versions of `Blob`s.
 *
 * @categoryDescription Declaration
 * `FileStream` can be referenced in schemas.
 *
 * ```ts
 * import { co, FileStream } from "jazz-tools";
 *
 * class MyCoMap extends CoMap {
 *   file = co.ref(FileStream);
 * }
 * ```
 *
 * @category CoValues
 */
export class FileStream extends CoValueBase {
    constructor(options) {
        super();
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            raw = rawOwner.createBinaryStream();
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _type: { value: "BinaryCoStream", enumerable: false },
            _raw: { value: raw, enumerable: false },
        });
    }
    static create(options) {
        return new this(options);
    }
    getChunks(options) {
        return this._raw.getBinaryChunks(options?.allowUnfinished);
    }
    isBinaryStreamEnded() {
        return this._raw.isBinaryStreamEnded();
    }
    start(options) {
        this._raw.startBinaryStream(options);
    }
    push(data) {
        this._raw.pushBinaryStreamChunk(data);
    }
    end() {
        this._raw.endBinaryStream();
    }
    toBlob(options) {
        const chunks = this.getChunks({
            allowUnfinished: options?.allowUnfinished,
        });
        if (!chunks) {
            return undefined;
        }
        // @ts-ignore
        return new Blob(chunks.chunks, { type: chunks.mimeType });
    }
    /**
     * Load a `FileStream` as a `Blob`
     *
     * @category Content
     */
    static async loadAsBlob(id, as, options) {
        let stream = await this.load(id, as, []);
        /**
         * If the user hasn't requested an incomplete blob and the
         * stream isn't complete wait for the stream download before progressing
         */
        if (!options?.allowUnfinished && !stream?.isBinaryStreamEnded()) {
            stream = await new Promise((resolve) => {
                const unsubscribe = subscribeToCoValue(this, id, as, [], (value) => {
                    if (value.isBinaryStreamEnded()) {
                        unsubscribe();
                        resolve(value);
                    }
                });
            });
        }
        return stream?.toBlob({
            allowUnfinished: options?.allowUnfinished,
        });
    }
    /**
     * Create a `FileStream` from a `Blob` or `File`
     *
     * @example
     * ```ts
     * import { co, FileStream } from "jazz-tools";
     *
     * const fileStream = await FileStream.createFromBlob(file, {owner: group})
     * ```
     * @category Content
     */
    static async createFromBlob(blob, options) {
        const stream = this.create({ owner: options.owner });
        const start = Date.now();
        const data = new Uint8Array(await blob.arrayBuffer());
        stream.start({
            mimeType: blob.type,
            totalSizeBytes: blob.size,
            fileName: blob instanceof File ? blob.name : undefined,
        });
        const chunkSize = MAX_RECOMMENDED_TX_SIZE;
        let lastProgressUpdate = Date.now();
        for (let idx = 0; idx < data.length; idx += chunkSize) {
            stream.push(data.slice(idx, idx + chunkSize));
            if (Date.now() - lastProgressUpdate > 100) {
                options.onProgress?.(idx / data.length);
                lastProgressUpdate = Date.now();
            }
            await new Promise((resolve) => setTimeout(resolve, 0));
        }
        stream.end();
        const end = Date.now();
        console.debug("Finished creating binary stream in", (end - start) / 1000, "s - Throughput in MB/s", (1000 * (blob.size / (end - start))) / (1024 * 1024));
        options.onProgress?.(1);
        return stream;
    }
    /**
     * Get a JSON representation of the `FileStream`
     * @category Content
     */
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
            ...this.getChunks(),
        };
    }
    /** @internal */
    [inspect]() {
        return this.toJSON();
    }
    /**
     * Load a `FileStream`
     * @category Subscription & Loading
     */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /**
     * Subscribe to a `FileStream`, when you have an ID but don't have a `FileStream` instance yet
     * @category Subscription & Loading
     */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /**
     * An instance method to subscribe to an existing `FileStream`
     * @category Subscription & Loading
     */
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
}
//# sourceMappingURL=coFeed.js.map